﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Core;
using HeuristicLab.Problems.BinPacking.Dimensions;
using HeuristicLab.Problems.BinPacking.PackingBin;
using HeuristicLab.Problems.BinPacking.PackingItem;
using HeuristicLab.Encodings.PackingEncoding.PackingPlan;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Common;
using HeuristicLab.Data;
using HeuristicLab.Collections;
using HeuristicLab.Problems.BinPacking.Interfaces;
using HeuristicLab.Encodings.PackingEncoding.GroupingVector;

namespace HeuristicLab.Problems.BinPacking.Decoders {
  [Item("Identical bin three dimensional direct grouping vector decoder", "<Description missing...>")]
  [StorableClass]
  public class DeepestBottomLeftGroupingVectorDecoder : PackingSolutionDecoder<
      ThreeDimensionalPacking,
      CuboidPackingBin,
      CuboidPackingItem> {

    public DeepestBottomLeftGroupingVectorDecoder()
      : base() {
        //EncodedSolutionParameter.ActualName = "EncodedSolution";
    }
    [StorableConstructor]
    protected DeepestBottomLeftGroupingVectorDecoder(bool deserializing) : base(deserializing) { }
    protected DeepestBottomLeftGroupingVectorDecoder(DeepestBottomLeftGroupingVectorDecoder original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new DeepestBottomLeftGroupingVectorDecoder(this, cloner);
    }


    public override PackingPlan<ThreeDimensionalPacking, CuboidPackingBin, CuboidPackingItem> CreatePackingPlanFromEncoding(IPackingSolutionEncoding encodedSolution, CuboidPackingBin binMeasures, ItemList<CuboidPackingItem> itemMeasures) {

      var solution = encodedSolution as GroupingVectorEncoding;
      if (solution == null) throw new InvalidOperationException("Encoding is not of type GroupingVector");

      PackingPlan<ThreeDimensionalPacking, CuboidPackingBin, CuboidPackingItem> result = new PackingPlan<ThreeDimensionalPacking, CuboidPackingBin, CuboidPackingItem>(
        PackingBinMeasuresParameter.ActualValue, PackingItemMeasuresParameter.ActualValue);

      ObservableDictionary<int, ThreeDimensionalPacking> packingPositions = DeepestBottomLeftFunctions.DeepestLeftBottomPacking(
        solution, 
        PackingBinMeasuresParameter.ActualValue, 
        PackingItemMeasuresParameter.ActualValue);


      result.PackingItemPositions = packingPositions;

      return result;
    }
  }
}
